from charms.layer import status
from charms.reactive import when, when_all, when_any, when_not, clear_flag, set_flag, hook, not_unless
from charmhelpers.core.hookenv import charm_dir, config, log
from utils import UnixUser

from textwrap import dedent

import glob
import os
import pygit2
import socket
import subprocess
import yaml


AUTOPKGTEST_LOCATION = os.path.expanduser('~ubuntu/autopkgtest')
AUTOPKGTEST_CLOUD_LOCATION = os.path.expanduser('~ubuntu/autopkgtest-cloud')
AUTODEP8_LOCATION = os.path.expanduser('~ubuntu/autodep8')

AUTOPKGTEST_CLONE_LOCATION = 'https://git.launchpad.net/~ubuntu-release/autopkgtest/+git/development'
AUTODEP8_CLONE_LOCATION = 'https://git.launchpad.net/~ubuntu-release/autopkgtest/+git/autodep8'

RABBITMQ_CRED_PATH = os.path.expanduser('~ubuntu/rabbitmq.cred')


@when_not('autopkgtest.autopkgtest_cloud_symlinked')
def symlink_autopkgtest_cloud():
    with UnixUser('ubuntu'):
        try:
            autopkgtest_cloud = os.path.join(charm_dir(), 'autopkgtest-cloud')
            os.symlink(autopkgtest_cloud, AUTOPKGTEST_CLOUD_LOCATION)
        except FileExistsError:
            pass
    set_flag('autopkgtest.autopkgtest_cloud_symlinked')


@when('apt.installed.git')
@when_not('autopkgtest.autodep8_cloned')
def clone_autodep8():
    status.maintenance('Cloning autodep8')
    log('Cloning autodep8', 'INFO')
    with UnixUser('ubuntu'):
        try:
            pygit2.clone_repository(AUTODEP8_CLONE_LOCATION,
                                    AUTODEP8_LOCATION)
            log('autodep8 cloned', 'INFO')
            status.active('autodep8 cloned')
            set_flag('autopkgtest.autodep8_cloned')
        except ValueError:
            log('autodep8 already cloned', 'INFO')


@when('autodep8_cloned')
def install_autodep8():
    pass


@when('apt.installed.git')
@when_not('autopkgtest.autopkgtest_cloned')
def clone_autopkgtest():
    status.maintenance('Cloning autopkgtest')
    log('Cloning autopkgtest', 'INFO')
    with UnixUser('ubuntu'):
        try:
            pygit2.clone_repository(AUTOPKGTEST_CLONE_LOCATION,
                                    AUTOPKGTEST_LOCATION)
            status.active('autopkgtest cloned')
            log('autopkgtest cloned', 'INFO')
            set_flag('autopkgtest.autopkgtest_cloned')
        except ValueError:
            # exists
            log('autopkgtest already cloned', 'INFO')


@when_all('autopkgtest.autopkgtest_cloud_symlinked',
          'autopkgtest.rabbitmq-configured')
def set_up_systemd_units():
    status.maintenance('Installing systemd units')
    for unit in glob.glob(os.path.join(charm_dir(), 'units', '*')):
        base = os.path.basename(unit)
        try:
            os.link(unit, os.path.join(os.path.sep,
                                       'etc',
                                       'systemd',
                                       'system',
                                       base))
            if '@' not in base:
                subprocess.check_call(['systemctl',
                                       'enable',
                                       base])
        except FileExistsError:
            pass

    subprocess.check_call(['systemctl',
                           'enable',
                           'autopkgtest.target'])
    subprocess.check_call(['systemctl',
                           'start',
                           'autopkgtest.target'])
    set_flag('autopkgtest.systemd_units_linked_and_enabled')
    status.active('systemd units installed and enabled')


@when_all('autopkgtest.systemd_units_linked_and_enabled',
          'autopkgtest.daemon-reload-needed')
@hook('upgrade-charm')
def daemon_reload():
    subprocess.check_call(['systemctl', 'daemon-reload'])
    clear_flag('autopkgtest.daemon-reload-needed')


@hook('install')
def enable_persistent_journal():
    try:
        journal_dir = os.path.join(os.path.sep, 'var', 'log', 'journal')
        os.makedirs(journal_dir)
        subprocess.check_call(['systemd-tmpfiles',
                               '--create',
                               '--prefix',
                               journal_dir])
        subprocess.check_call(['systemctl',
                               'restart',
                               'systemd-journald'])
    except FileExistsError:
        pass


@when('amqp.connected')
@when_not('amqp.available')
def setup_rabbitmq(rabbitmq):
    rabbitmq.request_access('autopkgtest-worker', '/')
    status.waiting('Waiting on RabbitMQ to configure vhost')


@when('amqp.available')
@when_not('autopkgtest.rabbitmq-configured')
def set_up_rabbitmq(rabbitmq):
    username = rabbitmq.username()
    password = rabbitmq.password()
    host = rabbitmq.private_address()
    status.maintenance('Configuring rabbitmq')
    log('Setting up rabbitmq connection to: {}@{}'.format(username, host))
    with open(RABBITMQ_CRED_PATH, 'w') as cred_file:
        cred_file.write(dedent('''\
            RABBIT_HOST="{host}"
            RABBIT_USER="{username}"
            RABBIT_PASSWORD="{password}"
            '''.format(**locals())))
    status.active('rabbitmq configured')
    set_flag('autopkgtest.rabbitmq-configured')


@when_not('amqp.available')
def clear_rabbitmq():
    try:
        log('rabbitmq not available, deleting credentials file')
        os.unlink(RABBITMQ_CRED_PATH)
        clear_flag('autopkgtest.rabbitmq-configured')
    except FileNotFoundError:
        pass


@when('config.changed.nova-rcs')
def update_nova_rcs():
    from tarfile import TarFile
    from io import BytesIO

    import base64

    rctar = config().get('nova-rcs')

    if not rctar:
        # set to the empty value
        return

    log('Extracting cloud .rc files...', 'INFO')

    clear_old_rcs()

    bytes_file = BytesIO(base64.b64decode(rctar))
    tar = TarFile(fileobj=bytes_file)

    log('...got {}'.format(', '.join(tar.getnames())), 'INFO')

    tar.extractall(os.path.expanduser('~ubuntu/cloudrcs/'))


@when('config.default.nova-rcs')
def clear_old_rcs():
    rcfiles = glob.glob(os.path.expanduser('~ubuntu/cloudrcs/*.rc'))

    if not rcfiles:
        return

    log('Deleting old cloud .rc files...', 'INFO')

    for rcfile in rcfiles:
        log('...deleting {}'.format(rcfile))
        os.unlink(rcfile)

    log('...done', 'INFO')


@when_all('autopkgtest.autopkgtest_cloud_symlinked',
          'autopkgtest.systemd_units_linked_and_enabled',
          'config.set.releases')
@when_any('config.set.n-workers',
          'config.set.lxd-remotes',
          'config.set.releases')
def enable_disable_units():
    from lib.systemd import set_up_systemd_units

    nworkers = config().get('n-workers') or ''
    lxdremotes = config().get('lxd-remotes') or ''
    releases = config().get('releases') or ''

    nworkers_yaml = yaml.load(nworkers, Loader=yaml.CSafeLoader)
    lxdremotes_yaml = yaml.load(lxdremotes, Loader=yaml.CSafeLoader)

    log('Setting up systemd worker units', 'INFO')
    status.maintenance('Enabling and starting worker units')
    set_up_systemd_units(nworkers_yaml or {},
                         lxdremotes_yaml or {},
                         releases.split())

    set_flag('autopkgtest.reload-needed')
    set_flag('autopkgtest.daemon-reload-needed')
    status.active('worker units up and running')


@when('apt.installed.lxd-client')
@when_any('config.set.n-workers',
          'config.set.lxd-remotes',
          'config.set.releases')
@when_not('autopkgtest.ubuntu_added_to_lxd_group')
def add_ubuntu_user_to_lxd_group():
    subprocess.check_call(['adduser',
                           'ubuntu',
                           'lxd'])
    set_flag('autopkgtest.ubuntu_added_to_lxd_group')


@when('config.changed.swift-auth-version')
def auth_version_changed():
    auth_version = config().get('swift-auth-version')

    if auth_version not in [2, 3]:
        raise ValueError('Only auth versions 2 and 3 are supported')

    if auth_version == 2:
        set_flag('autopkgtest.v2_auth')
        clear_flag('autopkgtest.v3_auth')
    else:
        clear_flag('autopkgtest.v2_auth')
        set_flag('autopkgtest.v3_auth')


@not_unless('autopkgtest.v3_auth')
@when_any('config.changed.swift-auth-url',
          'config.changed.swift-username',
          'config.changed.swift-password',
          'config.changed.swift-region',
          'config.changed.swift-project-domain-name',
          'config.changed.swift-project-name',
          'config.changed.swift-user-domain-name',
          'config.changed.swift-auth-version')
@when_all('config.set.swift-auth-url',
          'config.set.swift-username',
          'config.set.swift-password',
          'config.set.swift-region',
          'config.set.swift-project-domain-name',
          'config.set.swift-project-name',
          'config.set.swift-user-domain-name',
          'config.set.swift-auth-version')
def write_v3_config():
    write_swift_config()


@not_unless('autopkgtest.v2_auth')
@when_any('config.changed.swift-auth-url',
          'config.changed.swift-username',
          'config.changed.swift-password',
          'config.changed.swift-region',
          'config.changed.swift-auth-version',
          'config.changed.swift-tenant')
@when_all('config.set.swift-auth-url',
          'config.set.swift-username',
          'config.set.swift-password',
          'config.set.swift-region',
          'config.set.swift-auth-version',
          'config.set.swift-tenant')
def write_v2_config():
    write_swift_config()


def write_swift_config():
    with open(os.path.expanduser('~ubuntu/swift-password.cred'), 'w') as swift_password_file:
            for key in config():
                if key.startswith('swift') and config()[key] is not None:
                    swift_password_file.write('{}="{}"\n'.format(key.upper().replace('-', '_'),
                                                                 str(config()[key]).strip()))


@when_any('config.changed.worker-big-packages',
          'config.changed.worker-long-tests',
          'config.changed.worker-blacklist',
          'config.changed.worker-default-flavor',
          'config.changed.worker-big-flavor',
          'config.changed.worker-args',
          'config.changed.worker-setup-command',
          'config.changed.worker-setup-command2',
          'config.changed.releases',
          'config.changed.n-workers',
          'config.changed.lxd-remotes',
          'config.changed.mirror')
@when_any('config.set.nova-rcs',
          'config.set.lxd-remotes')
def write_worker_config():
    import configparser

    nworkers = config().get('n-workers') or ''
    lxdremotes = config().get('lxd-remotes') or ''

    if not nworkers and not lxdremotes:
        return

    nworkers_yaml = yaml.load(nworkers, Loader=yaml.CSafeLoader) or {}
    lxdremotes_yaml = yaml.load(lxdremotes, Loader=yaml.CSafeLoader) or {}

    conf = {
        'autopkgtest': {
            'checkout_dir': AUTOPKGTEST_LOCATION,
            'releases': config().get('releases'),
            'setup_command': config().get('worker-setup-command'),
            'setup_command2': config().get('worker-setup-command2'),
            'big_packages': config().get('worker-big-packages'),
            'long_tests': config().get('worker-long-tests'),
            'blacklist': config().get('worker-blacklist')
        },
        'virt': {
            'package_size_default': config().get('worker-default-flavor'),
            'package_size_big': config().get('worker-big-flavor'),
            'args': config().get('worker-args'),
        }
    }

    def subst(s):
        replacements = {
            '/CHECKOUTDIR/': AUTOPKGTEST_LOCATION,
            '/HOSTNAME/': socket.gethostname(),
            '/AUTOPKGTEST_CLOUD_DIR/': AUTOPKGTEST_CLOUD_LOCATION,
            '/MIRROR/': config().get('mirror'),
            '/NET_NAME/': config().get('net-name')
        }

        for k in replacements:
            if replacements[k]:
                s = s.replace(k, replacements[k])

        return s

    # prune anything which wasn't set (nested dict)
    conf = {k: {l: subst(w) for l, w in v.items() if w is not None} for k, v in conf.items()}

    def write(conf_file):
        with open(conf_file, 'w') as cf:
            cp = configparser.ConfigParser()
            cp.read_dict(conf)
            cp.write(cf)

    for region in nworkers_yaml:
        for arch in nworkers_yaml[region]:
            if arch == 'amd64':
                conf_file = os.path.join(os.path.expanduser('~ubuntu'), 'worker-{}.conf'.format(region))
                conf['autopkgtest']['architectures'] = 'amd64 i386'
                write(conf_file)
                break
            else:
                conf_file = os.path.join(os.path.expanduser('~ubuntu'), 'worker-{}-{}.conf'.format(region, arch))
                conf['autopkgtest']['architectures'] = arch
                write(conf_file)

    for arch in lxdremotes_yaml:
        conf_file = os.path.join(os.path.expanduser('~ubuntu'), 'worker-lxd-{}.conf'.format(arch))
        conf['autopkgtest']['architectures'] = arch
        write(conf_file)

    set_flag('autopkgtest.daemon-reload-needed')
    set_flag('autopkgtest.reload-needed')


@when('config.changed.net-name')
def write_net_name():
    clear_flag('autopkgtest.net-name-written')
    with open(os.path.expanduser('~ubuntu/net-name.rc'), 'w') as f:
        f.write('NET_NAME="{}"\n'.format(config().get('net-name')))
    set_flag('autopkgtest.net-name-written')
    set_flag('autopkgtest.reload-needed')


@when('config.changed.mirror')
def write_mirror():
    with open(os.path.expanduser('~ubuntu/mirror.rc'), 'w') as f:
        f.write('MIRROR="{}"\n'.format(config().get('mirror')))
    set_flag('autopkgtest.reload-needed')


@when('autopkgtest.reload-needed')
@when_not('autopkgtest.daemon-reload-needed')
def reload_systemd_units():
    from lib.systemd import reload_units
    reload_units()
    clear_flag('autopkgtest.reload-needed')
